// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information
use crate::{
    bundle::{Bundle, BundleType},
    firrtl::{self, ExportOptions},
    intern::Interned,
    module::Module,
    util::streaming_read_utf8::streaming_read_utf8,
};
use clap::{
    builder::{OsStringValueParser, TypedValueParser},
    Parser, Subcommand, ValueEnum, ValueHint,
};
use eyre::{eyre, Report};
use std::{
    error,
    ffi::OsString,
    fmt::{self, Write},
    fs, io, mem,
    num::NonZeroUsize,
    path::{Path, PathBuf},
    process,
    sync::{Condvar, Mutex, OnceLock},
};
use tempfile::TempDir;

pub type Result<T = (), E = CliError> = std::result::Result<T, E>;

pub struct CliError(Report);

impl fmt::Debug for CliError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.0.fmt(f)
    }
}

impl fmt::Display for CliError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.0.fmt(f)
    }
}

impl error::Error for CliError {}

impl From<io::Error> for CliError {
    fn from(value: io::Error) -> Self {
        CliError(Report::new(value))
    }
}

#[derive(Debug)]
pub struct AcquiredJob {
    job_count: NonZeroUsize,
}

impl Drop for AcquiredJob {
    fn drop(&mut self) {
        Self::change_job_count(Some(self.job_count), None);
    }
}

impl AcquiredJob {
    pub fn max_available_job_count() -> NonZeroUsize {
        static RETVAL: OnceLock<NonZeroUsize> = OnceLock::new();
        *RETVAL.get_or_init(|| {
            std::thread::available_parallelism().unwrap_or(NonZeroUsize::new(1).unwrap())
        })
    }
    fn change_job_count(released: Option<NonZeroUsize>, acquired: Option<NonZeroUsize>) {
        static AVAILABLE_JOB_COUNT: OnceLock<Mutex<usize>> = OnceLock::new();
        static COND_VAR: Condvar = Condvar::new();
        let mut available_job_count_lock = AVAILABLE_JOB_COUNT
            .get_or_init(|| Mutex::new(Self::max_available_job_count().get()))
            .lock()
            .unwrap();
        if let Some(released) = released {
            *available_job_count_lock = available_job_count_lock
                .checked_add(released.get())
                .expect("tried to release too many jobs");
            COND_VAR.notify_all();
        }
        if let Some(acquired) = acquired {
            loop {
                match available_job_count_lock.checked_sub(acquired.get()) {
                    Some(jobs_left) => {
                        *available_job_count_lock = jobs_left;
                        break;
                    }
                    None => {
                        available_job_count_lock = COND_VAR.wait(available_job_count_lock).unwrap()
                    }
                }
            }
        }
    }
    pub fn job_count(&self) -> NonZeroUsize {
        self.job_count
    }
    pub fn increase_job_count<R>(
        &mut self,
        new_minimum_count: NonZeroUsize,
        f: impl FnOnce(&mut AcquiredJob) -> R,
    ) -> R {
        if new_minimum_count <= self.job_count {
            return f(self);
        }
        struct ReleaseOnDrop<'a> {
            acquired_job: &'a mut AcquiredJob,
            old_job_count: NonZeroUsize,
        }
        impl Drop for ReleaseOnDrop<'_> {
            fn drop(&mut self) {
                AcquiredJob::change_job_count(
                    NonZeroUsize::new(self.acquired_job.job_count.get() - self.old_job_count.get()),
                    None,
                );
                self.acquired_job.job_count = self.old_job_count;
            }
        }
        let release_on_drop = ReleaseOnDrop {
            old_job_count: self.job_count,
            acquired_job: self,
        };
        // release our current jobs when acquiring new jobs to avoid deadlock
        Self::change_job_count(Some(release_on_drop.old_job_count), Some(new_minimum_count));
        release_on_drop.acquired_job.job_count = new_minimum_count;
        let retval = f(release_on_drop.acquired_job);
        drop(release_on_drop);
        retval
    }
    pub fn acquire_jobs(job_count: NonZeroUsize) -> Self {
        Self::change_job_count(None, Some(job_count));
        Self { job_count }
    }
    pub fn run_command<R>(
        &mut self,
        mut cmd: std::process::Command,
        f: impl FnOnce(&mut std::process::Command) -> std::io::Result<R>,
    ) -> std::io::Result<R> {
        // TODO: if we implement a make job server, add the proper env vars to cmd
        f(&mut cmd)
    }
}

pub trait RunPhase<Arg> {
    type Output;
    fn run(&self, arg: Arg) -> Result<Self::Output> {
        self.run_with_job(
            arg,
            &mut AcquiredJob::acquire_jobs(NonZeroUsize::new(1).unwrap()),
        )
    }
    fn run_with_job(&self, arg: Arg, acquired_job: &mut AcquiredJob) -> Result<Self::Output>;
}

#[derive(Parser, Debug, Clone)]
#[non_exhaustive]
pub struct BaseArgs {
    /// the directory to put the generated main output file and associated files in
    #[arg(short, long, value_hint = ValueHint::DirPath, required = true)]
    pub output: Option<PathBuf>,
    /// the stem of the generated main output file, e.g. to get foo.v, pass --file-stem=foo
    #[arg(long)]
    pub file_stem: Option<String>,
    #[arg(long, env = "FAYALITE_KEEP_TEMP_DIR")]
    pub keep_temp_dir: bool,
    #[arg(skip = false)]
    pub redirect_output_for_rust_test: bool,
}

impl BaseArgs {
    fn make_firrtl_file_backend(&self) -> Result<(firrtl::FileBackend, Option<TempDir>)> {
        let (dir_path, temp_dir) = match &self.output {
            Some(output) => (output.clone(), None),
            None => {
                let temp_dir = TempDir::new()?;
                if self.keep_temp_dir {
                    let temp_dir = temp_dir.into_path();
                    println!("created temporary directory: {}", temp_dir.display());
                    (temp_dir, None)
                } else {
                    (temp_dir.path().to_path_buf(), Some(temp_dir))
                }
            }
        };
        Ok((
            firrtl::FileBackend {
                dir_path,
                top_fir_file_stem: self.file_stem.clone(),
                circuit_name: None,
            },
            temp_dir,
        ))
    }
    /// handles possibly redirecting the command's output for Rust tests
    pub fn run_external_command(
        &self,
        _acquired_job: &mut AcquiredJob,
        mut command: process::Command,
    ) -> io::Result<process::ExitStatus> {
        if self.redirect_output_for_rust_test {
            let (reader, writer) = os_pipe::pipe()?;
            let mut reader = io::BufReader::new(reader);
            command.stderr(writer.try_clone()?);
            command.stdout(writer); // must not leave writer around after spawning child
            command.stdin(process::Stdio::null());
            let mut child = command.spawn()?;
            drop(command); // close writers
            Ok(loop {
                let status = child.try_wait()?;
                streaming_read_utf8(&mut reader, |s| {
                    // use print! so output goes to Rust test output capture
                    print!("{s}");
                    io::Result::Ok(())
                })?;
                if let Some(status) = status {
                    break status;
                }
            })
        } else {
            command.status()
        }
    }
}

#[derive(Parser, Debug, Clone)]
#[non_exhaustive]
pub struct FirrtlArgs {
    #[command(flatten)]
    pub base: BaseArgs,
    #[command(flatten)]
    pub export_options: ExportOptions,
}

#[derive(Debug)]
#[non_exhaustive]
pub struct FirrtlOutput {
    pub file_stem: String,
    pub top_module: String,
    pub output_dir: PathBuf,
    pub temp_dir: Option<TempDir>,
}

impl FirrtlOutput {
    pub fn file_with_ext(&self, ext: &str) -> PathBuf {
        let mut retval = self.output_dir.join(&self.file_stem);
        retval.set_extension(ext);
        retval
    }
    pub fn firrtl_file(&self) -> PathBuf {
        self.file_with_ext("fir")
    }
}

impl FirrtlArgs {
    fn run_impl(
        &self,
        top_module: Module<Bundle>,
        _acquired_job: &mut AcquiredJob,
    ) -> Result<FirrtlOutput> {
        let (file_backend, temp_dir) = self.base.make_firrtl_file_backend()?;
        let firrtl::FileBackend {
            top_fir_file_stem,
            circuit_name,
            dir_path,
        } = firrtl::export(file_backend, &top_module, self.export_options)?;
        Ok(FirrtlOutput {
            file_stem: top_fir_file_stem.expect(
                "export is known to set the file stem from the circuit name if not provided",
            ),
            top_module: circuit_name.expect("export is known to set the circuit name"),
            output_dir: dir_path,
            temp_dir,
        })
    }
}

impl<T: BundleType> RunPhase<Module<T>> for FirrtlArgs {
    type Output = FirrtlOutput;
    fn run_with_job(
        &self,
        top_module: Module<T>,
        acquired_job: &mut AcquiredJob,
    ) -> Result<Self::Output> {
        self.run_impl(top_module.canonical(), acquired_job)
    }
}

impl<T: BundleType> RunPhase<Interned<Module<T>>> for FirrtlArgs {
    type Output = FirrtlOutput;
    fn run_with_job(
        &self,
        top_module: Interned<Module<T>>,
        acquired_job: &mut AcquiredJob,
    ) -> Result<Self::Output> {
        self.run_with_job(*top_module, acquired_job)
    }
}

/// based on [LLVM Circt's recommended lowering options
/// ](https://circt.llvm.org/docs/VerilogGeneration/#recommended-loweringoptions-by-target)
#[derive(ValueEnum, Copy, Clone, Debug, PartialEq, Eq, Hash)]
#[non_exhaustive]
pub enum VerilogDialect {
    Questa,
    Spyglass,
    Verilator,
    Vivado,
    Yosys,
}

impl fmt::Display for VerilogDialect {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.write_str(self.as_str())
    }
}

impl VerilogDialect {
    pub fn as_str(self) -> &'static str {
        match self {
            VerilogDialect::Questa => "questa",
            VerilogDialect::Spyglass => "spyglass",
            VerilogDialect::Verilator => "verilator",
            VerilogDialect::Vivado => "vivado",
            VerilogDialect::Yosys => "yosys",
        }
    }
    pub fn firtool_extra_args(self) -> &'static [&'static str] {
        match self {
            VerilogDialect::Questa => &["--lowering-options=emitWireInPorts"],
            VerilogDialect::Spyglass => {
                &["--lowering-options=explicitBitcast,disallowExpressionInliningInPorts"]
            }
            VerilogDialect::Verilator => &[
                "--lowering-options=locationInfoStyle=wrapInAtSquareBracket,disallowLocalVariables",
            ],
            VerilogDialect::Vivado => &["--lowering-options=mitigateVivadoArrayIndexConstPropBug"],
            VerilogDialect::Yosys => {
                &["--lowering-options=disallowLocalVariables,disallowPackedArrays"]
            }
        }
    }
}

#[derive(Parser, Debug, Clone)]
#[non_exhaustive]
pub struct VerilogArgs {
    #[command(flatten)]
    pub firrtl: FirrtlArgs,
    #[arg(
        long,
        default_value = "firtool",
        env = "FIRTOOL",
        value_hint = ValueHint::CommandName,
        value_parser = OsStringValueParser::new().try_map(which::which)
    )]
    pub firtool: PathBuf,
    #[arg(long)]
    pub firtool_extra_args: Vec<OsString>,
    /// adapt the generated Verilog for a particular toolchain
    #[arg(long)]
    pub verilog_dialect: Option<VerilogDialect>,
    #[arg(long, short = 'g')]
    pub debug: bool,
}

#[derive(Debug)]
#[non_exhaustive]
pub struct VerilogOutput {
    pub firrtl: FirrtlOutput,
    pub verilog_files: Vec<PathBuf>,
}

impl VerilogOutput {
    pub fn main_verilog_file(&self) -> PathBuf {
        self.firrtl.file_with_ext("v")
    }
    fn unadjusted_verilog_file(&self) -> PathBuf {
        self.firrtl.file_with_ext("unadjusted.v")
    }
}

impl VerilogArgs {
    fn process_unadjusted_verilog_file(&self, mut output: VerilogOutput) -> Result<VerilogOutput> {
        let input = fs::read_to_string(output.unadjusted_verilog_file())?;
        let file_separator_prefix = "\n// ----- 8< ----- FILE \"";
        let file_separator_suffix = "\" ----- 8< -----\n\n";
        let mut input = &*input;
        let main_verilog_file = output.main_verilog_file();
        let mut file_name: Option<&Path> = Some(&main_verilog_file);
        loop {
            let (chunk, next_file_name) = if let Some((chunk, rest)) =
                input.split_once(file_separator_prefix)
            {
                let Some((next_file_name, rest)) = rest.split_once(file_separator_suffix) else {
                    return Err(CliError(eyre!("parsing firtool's output failed: found {file_separator_prefix:?} but no {file_separator_suffix:?}")));
                };
                input = rest;
                (chunk, Some(next_file_name.as_ref()))
            } else {
                (mem::take(&mut input), None)
            };
            let Some(file_name) = mem::replace(&mut file_name, next_file_name) else {
                break;
            };
            let file_name = output.firrtl.output_dir.join(file_name);
            fs::write(&file_name, chunk)?;
            if let Some(extension) = file_name.extension() {
                if extension == "v" || extension == "sv" {
                    output.verilog_files.push(file_name);
                }
            }
        }
        Ok(output)
    }
    fn run_impl(
        &self,
        firrtl_output: FirrtlOutput,
        acquired_job: &mut AcquiredJob,
    ) -> Result<VerilogOutput> {
        let Self {
            firrtl,
            firtool,
            firtool_extra_args,
            verilog_dialect,
            debug,
        } = self;
        let output = VerilogOutput {
            firrtl: firrtl_output,
            verilog_files: vec![],
        };
        let mut cmd = process::Command::new(firtool);
        cmd.arg(output.firrtl.firrtl_file());
        cmd.arg("-o");
        cmd.arg(output.unadjusted_verilog_file());
        if *debug {
            cmd.arg("-g");
            cmd.arg("--preserve-values=all");
        }
        if let Some(dialect) = verilog_dialect {
            cmd.args(dialect.firtool_extra_args());
        }
        cmd.args(firtool_extra_args);
        cmd.current_dir(&output.firrtl.output_dir);
        let status = firrtl.base.run_external_command(acquired_job, cmd)?;
        if status.success() {
            self.process_unadjusted_verilog_file(output)
        } else {
            Err(CliError(eyre!(
                "running {} failed: {status}",
                self.firtool.display()
            )))
        }
    }
}

impl<Arg> RunPhase<Arg> for VerilogArgs
where
    FirrtlArgs: RunPhase<Arg, Output = FirrtlOutput>,
{
    type Output = VerilogOutput;
    fn run_with_job(&self, arg: Arg, acquired_job: &mut AcquiredJob) -> Result<Self::Output> {
        let firrtl_output = self.firrtl.run_with_job(arg, acquired_job)?;
        self.run_impl(firrtl_output, acquired_job)
    }
}

#[derive(ValueEnum, Copy, Clone, Debug, PartialEq, Eq, Hash, Default)]
#[non_exhaustive]
pub enum FormalMode {
    #[default]
    BMC,
    Prove,
    Live,
    Cover,
}

impl FormalMode {
    pub fn as_str(self) -> &'static str {
        match self {
            FormalMode::BMC => "bmc",
            FormalMode::Prove => "prove",
            FormalMode::Live => "live",
            FormalMode::Cover => "cover",
        }
    }
    fn needs_extra_job(self) -> bool {
        match self {
            FormalMode::BMC => false,
            FormalMode::Prove => true,
            FormalMode::Live => false,
            FormalMode::Cover => false,
        }
    }
}

impl fmt::Display for FormalMode {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.write_str(self.as_str())
    }
}

#[derive(Clone)]
struct FormalAdjustArgs;

impl clap::FromArgMatches for FormalAdjustArgs {
    fn from_arg_matches(_matches: &clap::ArgMatches) -> Result<Self, clap::Error> {
        Ok(Self)
    }

    fn update_from_arg_matches(&mut self, _matches: &clap::ArgMatches) -> Result<(), clap::Error> {
        Ok(())
    }
}

impl clap::Args for FormalAdjustArgs {
    fn augment_args(cmd: clap::Command) -> clap::Command {
        cmd.mut_arg("output", |arg| arg.required(false))
            .mut_arg("verilog_dialect", |arg| {
                arg.default_value(VerilogDialect::Yosys.to_string())
                    .hide(true)
            })
    }

    fn augment_args_for_update(cmd: clap::Command) -> clap::Command {
        Self::augment_args(cmd)
    }
}

#[derive(Parser, Clone)]
#[non_exhaustive]
pub struct FormalArgs {
    #[command(flatten)]
    pub verilog: VerilogArgs,
    #[arg(
        long,
        default_value = "sby",
        env = "SBY",
        value_hint = ValueHint::CommandName,
        value_parser = OsStringValueParser::new().try_map(which::which)
    )]
    pub sby: PathBuf,
    #[arg(long)]
    pub sby_extra_args: Vec<OsString>,
    #[arg(long, default_value_t)]
    pub mode: FormalMode,
    #[arg(long, default_value_t = Self::DEFAULT_DEPTH)]
    pub depth: u64,
    #[arg(long, default_value = "z3")]
    pub solver: String,
    #[arg(long)]
    pub smtbmc_extra_args: Vec<String>,
    #[command(flatten)]
    _formal_adjust_args: FormalAdjustArgs,
}

impl fmt::Debug for FormalArgs {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let Self {
            verilog,
            sby,
            sby_extra_args,
            mode,
            depth,
            solver,
            smtbmc_extra_args,
            _formal_adjust_args: _,
        } = self;
        f.debug_struct("FormalArgs")
            .field("verilog", &verilog)
            .field("sby", &sby)
            .field("sby_extra_args", &sby_extra_args)
            .field("mode", &mode)
            .field("depth", &depth)
            .field("solver", &solver)
            .field("smtbmc_extra_args", &smtbmc_extra_args)
            .finish_non_exhaustive()
    }
}

impl FormalArgs {
    pub const DEFAULT_DEPTH: u64 = 20;
}

#[derive(Debug)]
#[non_exhaustive]
pub struct FormalOutput {
    pub verilog: VerilogOutput,
}

impl FormalOutput {
    pub fn sby_file(&self) -> PathBuf {
        self.verilog.firrtl.file_with_ext("sby")
    }
}

impl FormalArgs {
    fn sby_contents(&self, output: &FormalOutput) -> Result<String> {
        let Self {
            verilog: _,
            sby: _,
            sby_extra_args: _,
            mode,
            depth,
            smtbmc_extra_args,
            solver,
            _formal_adjust_args: _,
        } = self;
        let smtbmc_options = smtbmc_extra_args.join(" ");
        let top_module = &output.verilog.firrtl.top_module;
        let mut retval = format!(
            "[options]\n\
            mode {mode}\n\
            depth {depth}\n\
            wait on\n\
            \n\
            [engines]\n\
            smtbmc {solver} -- -- {smtbmc_options}\n\
            \n\
            [script]\n"
        );
        for verilog_file in &output.verilog.verilog_files {
            let verilog_file = verilog_file
                .to_str()
                .ok_or_else(|| CliError(eyre!("verilog file path is not UTF-8")))?;
            if verilog_file.contains(|ch: char| {
                (ch != ' ' && ch != '\t' && ch.is_ascii_whitespace()) || ch == '"'
            }) {
                return Err(CliError(eyre!(
                    "verilog file path contains characters that aren't permitted"
                )));
            }
            writeln!(retval, "read_verilog -sv -formal \"{verilog_file}\"").unwrap();
        }
        writeln!(retval, "prep -top {top_module}").unwrap();
        Ok(retval)
    }
    fn run_impl(
        &self,
        verilog_output: VerilogOutput,
        acquired_job: &mut AcquiredJob,
    ) -> Result<FormalOutput> {
        let output = FormalOutput {
            verilog: verilog_output,
        };
        let sby_file = output.sby_file();
        std::fs::write(&sby_file, self.sby_contents(&output)?)?;
        let mut cmd = process::Command::new(&self.sby);
        cmd.arg("-f");
        cmd.arg(sby_file.file_name().unwrap());
        cmd.args(&self.sby_extra_args);
        cmd.current_dir(&output.verilog.firrtl.output_dir);
        let new_minimum_count = if self.mode.needs_extra_job() {
            NonZeroUsize::new(2).unwrap()
        } else {
            NonZeroUsize::new(1).unwrap()
        };
        let new_minimum_count = AcquiredJob::max_available_job_count().min(new_minimum_count);
        let status = acquired_job.increase_job_count(new_minimum_count, |acquired_job| {
            self.verilog
                .firrtl
                .base
                .run_external_command(acquired_job, cmd)
        })?;
        if status.success() {
            Ok(output)
        } else {
            Err(CliError(eyre!(
                "running {} failed: {status}",
                self.sby.display()
            )))
        }
    }
}

impl<Arg> RunPhase<Arg> for FormalArgs
where
    VerilogArgs: RunPhase<Arg, Output = VerilogOutput>,
{
    type Output = FormalOutput;
    fn run_with_job(&self, arg: Arg, acquired_job: &mut AcquiredJob) -> Result<Self::Output> {
        let verilog_output = self.verilog.run_with_job(arg, acquired_job)?;
        self.run_impl(verilog_output, acquired_job)
    }
}

#[derive(Subcommand, Debug)]
enum CliCommand {
    /// Generate FIRRTL
    Firrtl(FirrtlArgs),
    /// Generate Verilog
    Verilog(VerilogArgs),
    /// Run a formal proof
    Formal(FormalArgs),
}

/// a simple CLI
///
/// Use like:
///
/// ```no_run
/// # use fayalite::prelude::*;
/// # #[hdl_module]
/// # fn my_module() {}
/// use fayalite::cli;
///
/// fn main() -> cli::Result {
///     cli::Cli::parse().run(my_module())
/// }
/// ```
///
/// You can also use it with a larger [`clap`]-based CLI like so:
///
/// ```no_run
/// # use fayalite::prelude::*;
/// # #[hdl_module]
/// # fn my_module() {}
/// use clap::{Subcommand, Parser};
/// use fayalite::cli;
///
/// #[derive(Subcommand)]
/// pub enum Cmd {
///     #[command(flatten)]
///     Fayalite(cli::Cli),
///     MySpecialCommand {
///         #[arg(long)]
///         foo: bool,
///     },
/// }
///
/// #[derive(Parser)]
/// pub struct Cli {
///     #[command(subcommand)]
///     cmd: Cmd, // or just use cli::Cli directly if you don't need more subcommands
/// }
///
/// fn main() -> cli::Result {
///     match Cli::parse().cmd {
///         Cmd::Fayalite(v) => v.run(my_module())?,
///         Cmd::MySpecialCommand { foo } => println!("special: foo={foo}"),
///     }
///     Ok(())
/// }
/// ```
#[derive(Parser, Debug)]
// clear things that would be crate-specific
#[command(name = "Fayalite Simple CLI", about = None, long_about = None)]
pub struct Cli {
    #[command(subcommand)]
    subcommand: CliCommand,
}

impl clap::Subcommand for Cli {
    fn augment_subcommands(cmd: clap::Command) -> clap::Command {
        CliCommand::augment_subcommands(cmd)
    }

    fn augment_subcommands_for_update(cmd: clap::Command) -> clap::Command {
        CliCommand::augment_subcommands_for_update(cmd)
    }

    fn has_subcommand(name: &str) -> bool {
        CliCommand::has_subcommand(name)
    }
}

impl<T> RunPhase<T> for Cli
where
    FirrtlArgs: RunPhase<T, Output = FirrtlOutput>,
{
    type Output = ();
    fn run_with_job(&self, arg: T, acquired_job: &mut AcquiredJob) -> Result<Self::Output> {
        match &self.subcommand {
            CliCommand::Firrtl(c) => {
                c.run_with_job(arg, acquired_job)?;
            }
            CliCommand::Verilog(c) => {
                c.run_with_job(arg, acquired_job)?;
            }
            CliCommand::Formal(c) => {
                c.run_with_job(arg, acquired_job)?;
            }
        }
        Ok(())
    }
}

impl Cli {
    /// forwards to [`clap::Parser::parse()`] so you don't have to import [`clap::Parser`]
    pub fn parse() -> Self {
        clap::Parser::parse()
    }
    /// forwards to [`RunPhase::run()`] so you don't have to import [`RunPhase`]
    pub fn run<T>(&self, top_module: T) -> Result<()>
    where
        Self: RunPhase<T, Output = ()>,
    {
        RunPhase::run(self, top_module)
    }
}
